//
// NETDEMO
// Copyright (c) 2003 Rama C. Hoetzlein
//


#include <conio.h>
#include <stdio.h>
#include <winsock2.h>
#include <time.h>

#include "netdemo-eng.h"

NetEng::NetEng ()
{
	mode = NET_NO_MODE;
	state = NET_OFF;
	max_net_sockets = 0;
	num_net_sockets = 0;
	net_sockets = NULL;
}

// Start Network Engine - Initialize winsock and enter NET_START state as client or server
void NetEng::Start (char m)
{	
	FD_ZERO (&socket_set);

	WSADATA WSAData;
	int status;
    if ((status = WSAStartup(MAKEWORD(1,1), &WSAData)) == 0) {
		printf ("NetEng: Started winsock...\n");
		mode = m;
		state = NET_START;
	} else {
		printf ("NetEng: Very bad! Unable to start up winsock\n");
		state = NET_OFF;
	}
}

void NetEng::Go (void)
{
	if (mode==NET_CLIENT) {
		switch (state) {
		case NET_OFF: break;
		case NET_START: ClientStart (); break;
		case NET_RESET: ClientReset (); break;
		case NET_SEARCHING: ClientSearching (); break;
		case NET_CONNECTING: ClientConnecting (); break;
		case NET_CONNECTED: ClientConnected (); break;
		}
	} else {
		switch (state) {
		case NET_OFF: break;
		case NET_START: ServerStart (); break;
		case NET_RESET: ServerReset (); break;
		case NET_PROCESSING: ServerProcessing (); break;
		}
	}
}

void NetEng::ClientStart (void)
{
	AddSocket (NET_UDP);
	data.CreateStream (NET_BUFSIZE);
	printf ("Client: (Start) Created a UDP socket.\n");
	state = NET_RESET;
	client_sn = 0;
}

void NetEng::ClientReset (void)
{	
	printf ("Client: (Reset) Reseting socket to broadcast.\n");
	SetSocketMode (0, NET_BROADCAST_ON);
	SetSocketDestBroadcast (0, NET_SEARCH_PORT);	
	SetSocketTimeout (0, 0, 100);
	SetSocketSendTime (0, clock());
	printf ("Client: (Reset) Broadcasting on address: %s\n", GetDestName(0));

	state = NET_SEARCHING;
}

void NetEng::ClientSearching (void)
{	
	char* buf;
	if (Recieve() > 0) {
		int len = data.GetData (&buf);
		if (strncmp (buf, "sFND", 4)==0) {
			printf ("Client: (Search) Server responded at: %s\n", GetSrcName(0));
			SetSocketSrcToDest (0, 0);
			SetSocketMode (0, NET_BROADCAST_OFF);
			SetSocketSendTime (0, clock()-CLOCKS_PER_SEC*3);
			state = NET_CONNECTING;
		}				
		//printf ("    Recv: %.*s \n", len, buf);
	} else if (clock() - GetSocketSendTime(0) > CLOCKS_PER_SEC) {
		printf ("Client: (Search) Searching...\n");
		printf ("Client: (Search) Broadcasting cSRCH to locate a server.\n");			
		Send (0, "cSRCH", 5);
	}
}

void NetEng::ClientConnecting (void)
{
	char* buf;
	if (Recieve() > 0) {
		int len = data.GetData (&buf);
		if (strncmp (buf, "sOK", 3)==0) {
			unsigned int ok_port = atoi (buf+3);
			
			SetSocketMode (0, NET_BROADCAST_OFF);	
			SetSocketSrcToDest (0, 0);					
			SetSocketDestPort (0, ok_port);
			SocketConnectToDest (0);
			
			printf ("Client: (Connecting) Server says OK to connect at %s\n", GetDestName(0));

			char sendbuf[20];		
			sprintf (sendbuf, "cCTD?%u", ok_port);		// Make sure we're connected
			Send (0, sendbuf, strlen(sendbuf));			

			printf ("Client: (Connecting) Sending cCTD? to confirm the connection\n");
		} else if (strncmp (buf, "sCTD", 4)==0) {			
			client_sn = atoi (buf+4);
			printf ("Client: (Connecting) Server says connected at %s on socket %d!\n", GetSrcName(0), client_sn);
			state = NET_CONNECTED;
			printf ("Client: (Connected) Connected...\n");
			printf ("Client: (Connected) Send/Recieve of data now possible.\n");
		} else if (strncmp (buf, "sNCTD", 5)==0) {
			printf ("Client: (Connecting) Server says were NOT connected.\n");
			state = NET_OFF;
		}
	} else if (clock() - GetSocketSendTime(0) > CLOCKS_PER_SEC*3) {
		printf ("Client: (Connecting) Connecting...\n");
		char disp[20];
		sprintf (disp, "cCNCT%s", "rama");
		Send (0, disp, strlen(disp));
		printf ("Client: (Connecting) Sending cCNCT to server to request connect\n");
	}
}

void NetEng::ClientConnected (void)
{
	Recieve ();

	int len;
	char* buf;
	
	while ((len=data.GetData(&buf)) > 0) {
		if (strncmp (buf, "cDAT", 4)==0) {
			from_who = atoi(buf+4);
			curr_byte = *(buf+5);
			byte_available = true;
		}
	}
}

void NetEng::ServerStart (void)
{
	AddSocket (NET_UDP);
	data.CreateStream (NET_BUFSIZE);
	printf ("Server: (Start) Created a UDP socket.\n");
	state = NET_RESET;
}

void NetEng::ServerReset (void)
{
	printf ("Server: (Reset) Reseting socket.\n");
	SetSocketMode (0, NET_BROADCAST_OFF);
	SetSocketSrcAny (0, NET_SEARCH_PORT);
	SocketBindToSrc (0);
	SetSocketTimeout (0, 0, 500);
	printf ("Server: (Reset) Set to recieve from ANY address on port %s\n", GetSrcName(0));

	serv_port = NET_SEARCH_PORT + 1;	
	state = NET_PROCESSING;
}	

void NetEng::ServerProcessing (void)
{
	char* buf;	
	if (Recieve() > 0) {					// Receive data on all sockets
		int len;
											// Process all incoming data
		while ((len=data.GetData(&buf)) > 0) {
		
		if (strncmp (buf, "cDAT", 4)==0) {			
			from_who = atoi(buf+4);
			curr_byte = *(buf+5);			
			byte_available = true;
		} else if (strncmp (buf, "cSRCH", 5)==0) {
			printf ("Server: (Processing) Client searching for server: %s\n", GetSrcName(0));
			SetSocketSrcToDest (0, 0);
			Send (0, "sFND", 4);			
			printf ("Server: (Processing) Sending sFND response to client..\n");
		} else if (strncmp (buf, "cCNCT", 5)==0) {
			printf ("Server: (Processing) Client requesting to connect: %s\n", GetSrcName(0));
			printf ("Server: (Processing) Permitting connect request.\n");
			SetSocketSrcToDest (0, 0);			
			int s = AddSocket (NET_UDP);			// Create new socket for client
			SetSocketSrcAny (s, serv_port);
			SetSocketSrcPort (s, serv_port);			
			SetSocketMode (s, NET_BROADCAST_OFF);	// Direct, no broadcasting. 
			SetSocketTimeout (s, 0, 500);
			SocketBindToSrc (s);					// Bind socket to this address & port only.
			printf ("Server: (Processing) Creating new client socket %d on port %d\n", s, serv_port);

			printf ("Server: (Processing) Sending sOK to client to grant connection.\n");
			char sendbuf[20];
			sprintf (sendbuf, "sOK%u", serv_port);	// Tell client it is connected.
			Send (0, sendbuf, strlen(sendbuf));		// WE ARE >NOT< YET SENDING ON NEW CLIENT SOCKET HERE
			serv_port++;
		} else if (strncmp (buf, "cCTD?", 5)==0) {
			unsigned int chk_port = atoi (buf+5);
			printf ("Server: (Processing) Client is asking if connected ok: %s\n", GetSrcName(0));
			int s = FindSocketWithSrcPort (chk_port);
			if (s > 0) {
				SetSocketSrcToDest (s, s);
				SocketConnectToDest (s);				

				printf ("Server: (Processing) Sending sCTD to client. Connected ok.\n");
				
				char sendbuf[20];
				sprintf (sendbuf, "sCTD%u", s);
				Send (s, sendbuf, strlen(sendbuf));
			} else {
				printf ("Server: (Processing) Sending sNCTD to client. Not connected ok.\n");

				Send (s, "sNCTD", 5);			
			}
		}
		
		}
	}	
}

int NetEng::AddSocket (char protocol)
{
	if (num_net_sockets==max_net_sockets) {
		if (max_net_sockets==0) max_net_sockets = 1; else max_net_sockets <<= 1;
		NetSock* new_list = new NetSock[max_net_sockets];
		if (net_sockets!=NULL) {
			memcpy (new_list, net_sockets, num_net_sockets*sizeof(NetSock));
			delete [] net_sockets;
		}
		net_sockets = new_list;
	}
	net_sockets[num_net_sockets].Create (protocol);
	num_net_sockets++;
	return num_net_sockets-1;
}

int	NetEng::FindSocketWithSrcPort (int chk_port)
{
	NetAddress d;	
	for (int n=0; n < num_net_sockets; n++) {
		d = GetSocketSrc (n);
		if (d.GetPort () == chk_port) return n;
	}
	return -1;
}

int NetEng::RecieveFrom (void)
{
	char buf[2000];
	int len = 2000;
	int n;
	int result, curr_result;

	FD_ZERO (&socket_set);
	for (n=0; n < num_net_sockets; n++) FD_SET (net_sockets[n].GetSocket(), &socket_set);

	result = select (1, &socket_set, NULL, NULL, net_sockets[0].GetTimeoutMdfy());
	if (result > 0) {
		result = 0;
		for (n=0; n < num_net_sockets; n++) {
			if (FD_ISSET (net_sockets[n].GetSocket(), &socket_set)) {				
				curr_result = net_sockets[n].RecvDataFrom (buf, len);
				if (curr_result > 0) {
					result += curr_result;
					data.AddData (buf, curr_result);
				}
			}
		}
		return result;
	}
	return 0;
}

int NetEng::Recieve (void)
{
	char buf[2000];
	int len = 2000;
	int n;
	int result, curr_result;

	FD_ZERO (&socket_set);
	for (n=0; n < num_net_sockets; n++) FD_SET (net_sockets[n].GetSocket(), &socket_set);

	result = select (1, &socket_set, NULL, NULL, net_sockets[0].GetTimeoutMdfy());
	if (result > 0) {
		result = 0;
		for (n=0; n < num_net_sockets; n++) {
			if (FD_ISSET (net_sockets[n].GetSocket(), &socket_set)) {
				if (net_sockets[n].GetConnected())	curr_result = net_sockets[n].RecvData (buf, len);
				else								curr_result = net_sockets[n].RecvDataFrom (buf, len);
				if (curr_result > 0) {
					result += curr_result;
					data.AddData (buf, curr_result);
				}
			}
		}
		return result;
	}
	return 0;
}

int NetEng::Send (int sn, char* buf, int len)
{
	if (GetSocketConnected(sn)) 	return SocketSendData (sn, buf, len);
	else							return SocketSendDataTo (sn, buf, len);
}

int NetEng::SendByteToServer (char c)
{
	char sendbuf[8];	
	sprintf (sendbuf, "cDAT%u%c", client_sn, c);
	Send (0, sendbuf, strlen(sendbuf));
	return 1;
}

int NetEng::SendByteToClients (char c)
{
	char sendbuf[8];	
	
	sprintf (sendbuf, "cDAT%u%c", client_sn, c);	
	for (int cn=1; cn < num_net_sockets; cn++) {
		Send (cn, sendbuf, strlen(sendbuf));
	}
	return 1;

}

int NetEng::RecieveByte (int* who)
{
	if (byte_available==true) {
		byte_available = false;
		*who = from_who;
		return (int) curr_byte;
	} else {
		return -1;
	}
}

char *NetEng::GetDestName (int sn)
{
	addrlen = 100;
	GetSocketDestName (sn, addrname, &addrlen);
	return addrname;
}

char *NetEng::GetSrcName (int sn)
{
	addrlen = 100;
	GetSocketSrcName (sn, addrname, &addrlen);
	return addrname;
}


